
 ---------------------------------------------------------------------
 --                                                                 --
 --                Infinite-Game Toolkit v. 1.0                     --
 --                                                                 --
 --            Routines For "Infinite" Game Elements                --
 --                                                                 --
 --             Created July 31, 2000 by Rod Jackson                --
 --             Reworked May 17, 2001 by Rod Jackson                --
 --                     All rights reserved.                        --
 --                                                                 --
 --                                                                 --
 -- This source code is made available free of charge for all       --
 -- public, private, and commercial use. It has been designed to    --
 -- run under version 2.0+ of the Euphoria interpreter by Rapid     --
 -- Development Software. Only authorized modifications or upgrades --
 -- should be distributed unless said modifications or upgrades are --
 -- clearly identified as "non-official" or "unofficial".           --
 --                                                                 --
 -- For more information on this library, please refer to the       --
 -- documentation files, or contact the author at:                  --
 --                                                                 --
 --    rodjackson@bigfoot.com                                       --
 --                                                                 --
 ---------------------------------------------------------------------

 --------------
 -- INCLUDES --
 --------------

include machine.e

 ---------------
 -- CONSTANTS --
 ---------------

constant T = 1 -- TRUE
constant F = 0 -- FALSE

constant VOWELS = "AEIOU " -- used for eliminating multiple consonants
constant THOUSAND = 1000 -- used for improving accuracy w/ probabilities

 ----------

global constant MAX_NUM_LETTERS_PER_NAME = 20 -- legal limit to name size

 -- these constants specify how you want
 -- names "cleaned" before being returned
global constant ALL_NAMES = 0
global constant NO_MULTIPLE_FIRST_LETTER = 1
global constant NO_MULTIPLE_CONSONANTS = 2
global constant NO_MULTIPLE_LETTERS = 3
global constant NO_THREE_PLUS_LETTERS = 4

 ---------------
 -- VARIABLES --
 ---------------

integer num_name_tables,          -- total number of namesets setup by IGT
        num_object_tables,        -- total number of objectsets setup by IGT
        prior_lower_limit,        -- previous lower limit given to igt_value()
        prior_upper_limit,        -- previous upper limit given to igt_value()
        offset,                   -- how far the lower limit is from 1
        true_upper,               -- the upper limit, adjusted by the offset
        num_names,                -- number of valid names to make the nameset
        L,                        -- used to hold lengths
        c,                        -- used to hold single ASCII characters
        iter,                     -- used to iterate through 'while' loops
        new_length,               -- length of a new name to construct
        first_letter,             -- first letter of a new name to construct
        last_letter,              -- prev. letter of new name
        next_letter,              -- next letter of new name
        was_double,               -- if we've already found a double-letter
        choice,                   -- actual choice returned from rand()
        letter_index,             -- an index to a letter table
        all_names,                -- was this indicator passed to igt_name()?
        no_multiple_first_letter, -- was this indicator passed to igt_name()?
        no_multiple_consonants,   -- was this indicator passed to igt_name()?
        no_multiple_letters,      -- was this indicator passed to igt_name()?
        no_three_plus_letters     -- was this indicator passed to igt_name()?

atom sum_values -- sum of a count of elements or probabilities

sequence next_letter_tables,  -- one table per nameset; prob. of any specific
                              --   letter following another in a name
         first_letter_tables, -- one table per nameset; prob. of any specific
                              --   letter being the first in a name
         length_tables,       -- one table per nameset; prob. of a name being
                              --   of any specific length
         object_tables,       -- one table per objectset; each table pairs
                              --   specific objects with the prob. of that
                              --   object being chosen from the set
         curr_next_letters,   -- a 2D table giving the probability of each
                              --   letter following each other letter
         curr_first_letters,  -- a table giving the probability of each
                              --   letter occuring first in a name
         curr_lengths,        -- a table giving the probability of each name
                              --   length, from 2 to MAX_NUM_LETTERS_PER_NAME
                              --   being the actual length of the name
         curr_objects,        -- a table that pairs specific objects to be
                              --   chosen with the probability of that object
                              --   actually being chosen
         new_name,            -- a new name constructed from a given nameset
         next_letters,        -- slice of a table of next letters; the prob.
                              --   of each letter following one specific one
         curr_name            -- the curr. base name from a given list

object o -- used to hold an element not known to be an integer

 -----------
 -- TYPES --
 -----------

 -- A valid name for basing a nameset on....
 -- Can't be longer than MAX_NUM_LETTERS_PER_NAME,
 -- must be at least 2 letters, must contain all
 -- integers, only uppercase letters and spaces.
type valid_basename (object x)

   if (sequence (x)) then
      L = length (x)
      if (L > 1) then
         if (L <= MAX_NUM_LETTERS_PER_NAME) then
            for i = 1 to L do
               o = x[i]
               if (not (integer (o))) then
                  return 0
               else
                  c = o
               end if
               if (c != ' ') and ((c > 'Z') or (c < 'A')) then
                  return 0
               end if
            end for
            return 1
         end if
      end if
   end if

   return 0

end type -- valid_basename()

 --------------
 -- ROUTINES --
 --------------

 ----------------------------------------------------------------------------
 -- trim_basename: removes the leading, trailing, and multiple spaces      --
 --    from the given basename, returning the result.                      --
 ----------------------------------------------------------------------------
function trim_basename (sequence b)

   L = length (b)

   while (L > 0) do
      if (b[1] = ' ')  then
         b = b[2..L]
         L = L - 1
      else
         L = 0
      end if
   end while

   L = length (b)
   while (L > 0) do
      if (b[L] = ' ') then
         L = L - 1
         b = b[1..L]
      else
         L = 0
      end if
   end while

   L = match ("  ", b)
   while (L > 0) do
      b = b[1..L] & b[L+2..length(b)]
      L = match ("  ", b)
   end while

   return b

end function -- trim_basename()

 --------------------
 -- INITIALIZATION --
 --------------------

next_letter_tables  = {}
first_letter_tables = {}
length_tables = {}
object_tables = {}
num_name_tables   = 0
num_object_tables = 0

prior_lower_limit = 1
prior_upper_limit = 1
offset = 0
true_upper = 1

all_names = T
no_multiple_first_letter = F
no_multiple_consonants   = F
no_multiple_letters      = F
no_three_plus_letters    = F

 --------------------------
 -- MAIN GLOBAL ROUTINES --
 --------------------------

 ----------------------------------------------------------------------------
 -- igt_values: given the integer seed and a sequence of sets of integer   --
 --    boundaries (lower and upper), this routine returns a sequence of    --
 --    sequential consistent (yet random) integer values within those      --
 --    boundaries (inclusive.)                                             --
 ----------------------------------------------------------------------------
global function igt_values (integer seed, sequence limit_list)

   sequence limits, results

   results = {}

   set_rand (seed)

   for i = 1 to length (limit_list) do

      limits = limit_list[i]

      if (length (limits) > 1) then
         prior_lower_limit = limits[1]
         prior_upper_limit = limits[2]
         offset = prior_lower_limit - 1
         true_upper = prior_upper_limit - offset
      end if

      results = results & (rand (true_upper) + offset)

   end for -- end loop through all value-limit sets

   return results

end function -- igt_values()

 ----------------------------------------------------------------------------
 -- new_igt_nameset: establishes a new nameset upon which a random name    --
 --    based on the characteristics of that nameset can be generated.      --
 --    These characteristics are determined by the names given to generate --
 --    the nameset.                                                        --
 ----------------------------------------------------------------------------
global function new_igt_nameset (sequence basenames)

   num_names = 0

   curr_lengths = repeat (0, MAX_NUM_LETTERS_PER_NAME)
   curr_first_letters = repeat (0, 26)
   curr_next_letters = repeat (repeat (0, 27), 27)

   for i = 1 to length (basenames) do
      curr_name = trim_basename (basenames[i])
      if (valid_basename (curr_name)) then
         num_names = num_names + 1
         curr_lengths[length(curr_name)] =
            curr_lengths[length(curr_name)] + 1
         letter_index = curr_name[1] - 'A' + 1
         curr_first_letters[letter_index] =
            curr_first_letters[letter_index] + 1
         for j = 2 to length (curr_name) do
            c = curr_name[j]
            if (c = ' ') then
               c = 27
            else
               c = c - 'A' + 1
            end if
            curr_next_letters[letter_index][c] =
               curr_next_letters[letter_index][c] + 1
            letter_index = c
         end for
      end if
   end for

    -- only continue if we have at least one valid basename
   if (num_names < 1) then
      return -1
   else
      curr_lengths = curr_lengths / num_names
      curr_first_letters = curr_first_letters / num_names
      for i = 1 to 27 do
         sum_values = 0
         for j = 1 to 27 do
            sum_values = sum_values + curr_next_letters[i][j]
         end for
         if (sum_values > 0) then
            curr_next_letters[i] = curr_next_letters[i] / sum_values
         end if
      end for
   end if

   num_name_tables = num_name_tables + 1
   length_tables = append (length_tables, curr_lengths)
   first_letter_tables = append (first_letter_tables, curr_first_letters)
   next_letter_tables = append (next_letter_tables, curr_next_letters)

   return num_name_tables

end function -- new_igt_nameset()

 ----------------------------------------------------------------------------
 -- igt_names:  returns randomly-generated names based on the given        --
 --    randomization seed and the given namesets. Flags can be set to      --
 --    specify how each name is to be "cleaned" before being returned      --
 --    (remove multiple consecutive consonants, etc.)                      --
 ----------------------------------------------------------------------------
global function igt_names (integer  seed,
                           sequence nameset_list,
                           sequence flags_list)

   integer  nameset
   sequence flags, results

   set_rand (seed)

   results = {}

   for iii = 1 to length (nameset_list) do

      nameset = nameset_list[iii]
      flags   = flags_list[iii]

      curr_lengths = length_tables[nameset]
      curr_first_letters = first_letter_tables[nameset]
      curr_next_letters = next_letter_tables[nameset]

      sum_values = 0
      new_length = rand (THOUSAND)
      for i = 2 to MAX_NUM_LETTERS_PER_NAME do
         sum_values = sum_values + (curr_lengths[i] * THOUSAND)
         if (new_length <= sum_values) then
            new_length = i
            exit
         end if
      end for

      if (new_length > MAX_NUM_LETTERS_PER_NAME) then
         new_length = MAX_NUM_LETTERS_PER_NAME
      end if

      new_name = repeat (' ', new_length)

       -- determine the first letter of the name
      first_letter = rand (THOUSAND)
      sum_values = 0
      for i = 1 to 26 do
         sum_values = sum_values + (curr_first_letters[i] * THOUSAND)
         if (first_letter <= sum_values) then
            first_letter = i
            exit
         end if
      end for

      if (first_letter > 26) then
         first_letter = 'Z'
      else
         first_letter = first_letter + 'A' - 1
      end if

      new_name[1] = first_letter

       -- now construct the rest of the name
      last_letter = first_letter - 'A' + 1
      for i = 2 to new_length do
         next_letter = rand (THOUSAND)
         sum_values = 0
         next_letters = curr_next_letters[last_letter]
         for j = 1 to 27 do
            sum_values = sum_values + (next_letters[j] * THOUSAND)
            if (next_letter <= sum_values) then
               next_letter = j
               exit
            end if
         end for
         if (next_letter > 27) then
            next_letter = 27
         end if
         if (next_letter < 27) then
            new_name[i] = next_letter + 'A' - 1
         end if
         last_letter = next_letter
      end for

       -- remove any trailing spaces
      new_name = trim_basename (new_name)

       -- establish the flags
      if (length (flags) > 0) then
         all_names = T
         no_multiple_first_letter = F
         no_multiple_consonants = F
         no_multiple_letters = F
         no_three_plus_letters = F
         if not (find (ALL_NAMES, flags)) then
            if (find (NO_MULTIPLE_FIRST_LETTER, flags)) then
               all_names = F
               no_multiple_first_letter = T
            end if
            if (find (NO_MULTIPLE_CONSONANTS, flags)) then
               all_names = F
               no_multiple_consonants = T
            end if
            if (find (NO_MULTIPLE_LETTERS, flags)) then
               all_names = F
               no_multiple_letters = T
               no_three_plus_letters = T
               no_multiple_consonants = T
               no_multiple_first_letter = T
            end if
            if (find (NO_THREE_PLUS_LETTERS, flags)) then
               all_names = F
               no_three_plus_letters = T
            end if
         end if
      end if

       -- now check the flags to see if we need to modify this name
      if not (all_names) then
          -- remove all multiple consecutive letters from the name
         if (no_multiple_letters) then
            last_letter = 0
            iter = 1
            while (iter <= length (new_name)) do
               next_letter = new_name[iter]
               if (next_letter = last_letter) then
                  new_name = new_name[1..iter-1] &
                     new_name[iter+1..length(new_name)]
               else
                  iter = iter + 1
                  last_letter = next_letter
               end if
            end while

         else
            if (no_multiple_consonants) then
               last_letter = 0
               iter = 1
               while (iter <= length (new_name)) do
                  next_letter = new_name[iter]
                  if (next_letter = last_letter) then
                     new_name = new_name[1..iter-1] &
                        new_name[iter+1..length(new_name)]
                  else
                     iter = iter + 1
                     if not (find (next_letter, VOWELS)) then
                        last_letter = next_letter
                     end if
                  end if
               end while
            end if

            if (no_multiple_first_letter) then
               last_letter = new_name[1]
               iter = 2
               while (iter <= length (new_name)) do
                  next_letter = new_name[iter]
                  if (next_letter = last_letter) then
                     new_name = new_name[1..iter-1] &
                        new_name[iter+1..length(new_name)]
                  else
                     iter = length (new_name) + 1
                  end if
               end while
            end if

            if (no_three_plus_letters) then
               last_letter = 0
               was_double = 0
               iter = 1
               while (iter <= length (new_name)) do
                  next_letter = new_name[iter]
                  if (next_letter = last_letter) then
                     if (was_double) then
                        new_name = new_name[1..iter-1] &
                           new_name[iter+1..length(new_name)]
                     else
                        was_double = 1
                        iter = iter + 1
                     end if
                  else
                     iter = iter + 1
                     last_letter = next_letter
                     was_double = 0
                  end if
               end while
            end if

         end if -- end check for multiple-letters flag
      end if -- end check for all-names flag

      results = append (results, new_name)

   end for -- end loop through all nameset/flag combos

   return results

end function -- igt_names()

 ----------------------------------------------------------------------------
 -- new_igt_objectset: establishes a new objectset upon which objects can  --
 --    be pulled from at random. It must be given a list of pairs: each    --
 --    pair contains an object available for selection in the set, and the --
 --    probability of that object being pulled. Probabilities should add   --
 --    to 1.                                                               --
 ----------------------------------------------------------------------------
global function new_igt_objectset (sequence probables)

   num_object_tables = num_object_tables + 1
   object_tables = append (object_tables, probables)

   return num_object_tables

end function -- new_igt_objectset()

 ----------------------------------------------------------------------------
 -- igt_objects: pulls and returns an object at random from the given      --
 --    objectset, based on the randomization seed, and the probabilities   --
 --    of each object actually being pulled.                               --
 ----------------------------------------------------------------------------
global function igt_objects (integer seed, sequence objectset_list)

   integer  objectset
   sequence results

   results = {}
   set_rand (seed)

   for iii = 1 to length (objectset_list) do

      objectset = objectset_list[iii]

      curr_objects = object_tables[objectset]

       -- determine the object to return
      choice = rand (THOUSAND)
      sum_values = 0
      for i = 1 to length (curr_objects) do
         sum_values = sum_values + (curr_objects[i][2] * THOUSAND)
         if (choice <= sum_values) then
            choice = i
            exit
         end if
      end for

      if (choice > length (curr_objects)) then
         choice = length (curr_objects)
      end if

      results = append (results, curr_objects[choice][1])

   end for -- loop through all object sets

   return results

end function -- igt_objects()

